"""\
Flash message service

Userd for storing a flash message in a cookie so that the result of an action
performed before an HTTP redirect can be displayed to the user on the page they
end up seeing.

The code doesn't use Python's cookie support (which always appeared
over-complex) and instead uses a simple URL encoding scheme that allows all
unicode characters to be safely encoded in the cookie.

The maximum length of the message is 4070 characters assuming you use the
default allowed message types and only the letters a-z, A-Z, 0-9 and _. Any
other characters (or message types with a name longer than 'good') will use
more space when encoded in the cookie so you won't get quite so many
characters. You can also use Unicode characters but again the message itself
has to be slightly shorter because of the encoding used.

To read a flash message you would do this:

::
    
    print flow.flash.message
    print flow.flash.message_type
    flow.flash.remove()

If no flash message is present, ``flow.flash.message`` and
``flow.flash.message_type`` will be ``None``. 

To set a flash message you'd do this:

::

    flow.flash.set('Hi there!', message_type='good')

If you call ``flow.flash.set()`` without a message_type, the first allowed
message type is used (which by default is ``'good'``).

Once the flash message is set, reading it will return the set messages, not the
original ones from the request. If you don't remove the flash message it will be
present on the next request too.

.. warning ::

    The messages you set will be stored client-side in the cookie so could be
    ready by any other application on the same domain or by anyone with access
    to the user's browser's cookie store. Don't store passwords or other 
    sensitive data in the flash message.

The module also contains a commented-out version of a Middleware component from
an earlier implementation, left in the source code on the off chance it might
be useful later on.

"""

import logging
import urllib
from bn import AttributeDict

log = logging.getLogger(__name__)

class FlashMessageError(Exception):
    pass

#def get_service_ensuring_free_key(environ, key):
#    if not environ.has_key('service.service'):
#        raise FlashMessageError(
#            'The PowerPack service is not present so the FlashMessage middleware '
#            'cannot function'
#        )
#    if environ['service.service'].has_key(key):
#        log.error('The %r key is already being used in the service', key)
#    return environ['service.service']
#
#
#def FlashMessageMiddleware(
#    app, 
#    service_key='flash'
#):
#    """\
#    This middleware relies on the powerpack.service middleware being in place
#    """
#    def flash_middleware_request_handler(environ, start_response):
#        # If the flash service isn't being used, this middleware doesn't need
#        # to do anything
#        if not environ['wsgiorg.service'].has_key(service_key):
#            return app(environ, start_response)
#
#        # We use lists so that they can be modified in the remove()
#        # and set() functions
#        #cookie_strings = []
#
#        
#        # Set up the service and functions
#        # service = get_service_ensuring_free_key(environ, service_key)
#
#        service = environ['wsgiorg.service']
#
#        # Populate service.flash
#        if environ.get('HTTP_COOKIE'):
#            for cookie_pair in environ.get('HTTP_COOKIE').split('; '):
#                parts = cookie_pair.split('=')
#                if parts[0] == service[service_key].cookie_name and parts[1]:
#                    result = decode_flash(parts[1])
#                    service[service_key]['message'] = result['message']
#                    service[service_key]['message_type'] = result['message_type']
#                    service[service_key]['cookie_strings'].append(cookie_pair)
#                    break
#
#        def flash_start_response(status, headers, exc_info=None):
#            start_response_called = service[service_key].start_response_called
#            start_response_called.append(True)
#            cookie_string = ''
#            cookie_strings = service[service_key].cookie_strings
#            if cookie_strings:
#                # Set the flash message
#                new_headers = []
#                for header in headers:
#                    if header[0].lower() == 'Set-Cookie':
#                        cookie_string = header[1]
#                    else:
#                        new_headers.append(header)
#                if cookie_string:
#                    new_headers.append(
#                        ('Set-Cookie', cookie_string+'; message='+cookie_strings[0])
#                    )
#                else:
#                    new_headers.append(('Set-Cookie', cookie_strings[0]+'; path=/;'))
#                log.debug('Flash message response cookie: %r', cookie_strings[0])
#            else:
#                new_headers = headers
#                log.debug('No flash message set')
#            return start_response(status, new_headers, exc_info)
#        return app(environ, flash_start_response)
#    return flash_middleware_request_handler

def encode_flash(message, message_type, allowed_message_types):
    """
    'message_type:good&message:asd%3D%2B%C3%A7'
    """
    if message_type not in allowed_message_types:
        raise FlashMessageError(
            "Expected message_type to be one of %s, not %r"%(
                str(allowed_message_types)[1:-1], 
                message_type
            )
        )
    result = urllib.urlencode(
        {
            'message_type': message_type, 
            'message': message.encode('utf8')
        }
    ).replace('=',':')
    if len(result) >= 4096:
        raise FlashMessageError(
            'The encoded flash message is too long to store in the cookie'
        )
    return result

def decode_flash(cookie_value):
    parts = cookie_value.split('&')
    result = {}
    for part in parts:
        k, v = part.split(':')
        result[k] = urllib.unquote(v.replace(':', '=').replace('+', ' ')).decode('utf8')
    return result

class FlashMessageService(object):
    def __init__(self, cookie_name='flash', allowed_message_types=['good', 'normal', 'error']):
        if not len(allowed_message_types):
            raise FlashMessageError(
                'You must have at least one allowed message type'
            )
        self.allowed_message_types = allowed_message_types
        self.cookie_name = cookie_name

    @staticmethod
    def config(flow, name):
        return {}

    @staticmethod
    def create(flow, name, config):
        return FlashMessageService()

    def start(self, flow, name):
        def remove():
            flow.flash['message'] = None
            flow.flash['message_type'] = None
            while(flow[name].cookie_strings):
                flow[name].cookie_strings.pop()
            flow[name].cookie_strings.append(flow[name].cookie_name+'=')

        def set(message, message_type=None):
            if message_type is None: 
                message_type = self.allowed_message_types[0]
            if flow[name]['start_response_called']:
                raise FlashMessageError(
                    'You can\'t set a flash message after '
                    'start_response() has been called'
                )
            flow[name]['message'] = message
            flow[name]['message_type'] = message_type
            flow[name]['cookie_strings'].append(
                flow[name].cookie_name+'='+encode_flash(
                    message,
                    message_type,
                    self.allowed_message_types
                )
            )
       
        flow[name] = AttributeDict(
            message=None, 
            message_type=None, 
            remove=remove,
            set=set,
            cookie_strings=[],
            start_response_called = [],
            cookie_name = self.cookie_name,
        )

        # Populate flow.flash
        if flow.http.environ.get('HTTP_COOKIE'):
            for cookie_pair in flow.http.environ.get('HTTP_COOKIE').split('; '):
                parts = cookie_pair.split('=')
                if parts[0] == flow[name].cookie_name and parts[1]:
                    result = decode_flash(parts[1])
                    flow[name]['message'] = result['message']
                    flow[name]['message_type'] = result['message_type']
                    flow[name]['cookie_strings'].append(cookie_pair)
                    break

    def stop(self, flow, name):
        cookie_string = ''
        cookie_strings = flow[name].cookie_strings
        if cookie_strings:
            # Set the flash message
            new_headers = []
            for header in flow.http.response.headers:
                if header[0].lower() == 'Set-Cookie':
                    cookie_string = header[1]
                else:
                    new_headers.append(header)
            if cookie_string:
                new_headers.append(
                    ('Set-Cookie', cookie_string+'; message='+cookie_strings[0])
                )
            else:
                new_headers.append(('Set-Cookie', cookie_strings[0]+'; path=/;'))
            log.debug('Flash message response cookie: %r', cookie_strings[0])
            flow.http.response.headers = new_headers
        else:
            log.debug('No flash message set')

